"""Bitso tap class."""

from __future__ import annotations

import structlog
from singer_sdk import Stream, Tap
from singer_sdk import typing as th

from tap_bitso import streams


def console_formatter(*, colors: bool = True) -> structlog.stdlib.ProcessorFormatter:
    """Return a console formatter for structlog.

    Args:
        colors: Whether to use colors in the console output.

    Returns:
        A structlog formatter.
    """
    return structlog.stdlib.ProcessorFormatter(
        processor=structlog.dev.ConsoleRenderer(colors=colors),
        foreign_pre_chain=[
            structlog.stdlib.add_log_level,
            structlog.processors.TimeStamper(fmt="iso"),
        ],
    )


def json_formatter() -> structlog.stdlib.ProcessorFormatter:
    """Return a JSON formatter for structlog.

    Returns:
        A structlog formatter.
    """
    return structlog.stdlib.ProcessorFormatter(
        processor=structlog.processors.JSONRenderer(),
        foreign_pre_chain=[
            structlog.stdlib.add_logger_name,
            structlog.stdlib.add_log_level,
            structlog.processors.TimeStamper(fmt="iso"),
            structlog.stdlib.PositionalArgumentsFormatter(),
        ],
    )


class TapBitso(Tap):
    """Bitso tap class."""

    name = "tap-bitso"

    config_jsonschema = th.PropertiesList(
        th.Property(
            "key",
            th.StringType,
            required=True,
            description="Bitso API Key",
        ),
        th.Property(
            "secret",
            th.StringType,
            required=True,
            description="Bitso API Secret",
        ),
        th.Property(
            "base_url",
            th.StringType,
            default="https://api.bitso.com",
            description="Bitso API base URL",
        ),
        th.Property(
            "books",
            th.ArrayType(th.StringType),
            default=["btc_mxn"],
            description=(
                "Specifies which book to use for `tickers` and other endpoints"
            ),
        ),
    ).to_dict()

    def discover_streams(self) -> list[Stream]:
        """Return a list of discovered streams.

        Returns:
            A list of streams.
        """
        return [
            streams.BooksStream(tap=self),
            # streams.LedgerStream(tap=self),  # Removed upstream?  # noqa: ERA001
            streams.TickersStream(tap=self),
            streams.TradesStream(tap=self),
            streams.UserTradesStream(tap=self),
        ]



"""Tests standard tap features using the built-in SDK tests library."""

from __future__ import annotations

from typing import Any

from singer_sdk.testing import SuiteConfig, get_tap_test_class

from tap_bitso.tap import TapBitso

SAMPLE_CONFIG: dict[str, Any] = {}


TestTapBitso = get_tap_test_class(
    TapBitso,
    config=SAMPLE_CONFIG,
    suite_config=SuiteConfig(
        ignore_no_records_for_streams=["user_trades"],
        max_records_limit=50,
    ),
)



"""REST client handling, including BitsoStream base class."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generator

import backoff
from singer_sdk.exceptions import RetriableAPIError
from singer_sdk.streams import RESTStream
from structlog.contextvars import bind_contextvars

from tap_bitso.auth import BitsoAuthenticator

if TYPE_CHECKING:
    import requests


class BitsoStream(RESTStream[str]):
    """Bitso stream class."""

    records_jsonpath = "$.payload[*]"
    book_based = False
    retry_codes = (400,)

    def get_records(
        self,
        context: dict[str, Any] | None,
    ) -> Generator[dict[str, Any], None, None]:
        """Return a generator of row-type dictionary objects.

        Each row emitted should be a dictionary of property names to their values.

        Args:
            context: Stream partition or context dictionary.

        Yields:
            One item per (possibly processed) record in the API.
        """
        bind_contextvars(context=context, stream=self.name)
        yield from super().get_records(context=context)

    @property
    def url_base(self) -> str:
        """Get base URL for the Bitso API from config.

        Returns:
            Base URL for all API requests.
        """
        return self.config["base_url"]  # type: ignore[no-any-return]

    @property
    def authenticator(self) -> BitsoAuthenticator:
        """Return a new authenticator object.

        Returns:
            The Bitso API authenticator object.
        """
        return BitsoAuthenticator.create_for_stream(self)

    @property
    def http_headers(self) -> dict[str, Any]:
        """Return the http headers needed.

        Returns:
            A mapping of HTTP headers.
        """
        headers = {}
        if "user_agent" in self.config:
            headers["User-Agent"] = self.config.get("user_agent")
        return headers

    def get_url_params(
        self,
        context: dict[str, Any] | None,
        next_page_token: str | None,
    ) -> dict[str, Any]:
        """Return a dictionary of values to be used in URL parameterization.

        Args:
            context: Stream sync context.
            next_page_token: Value used to retrieve the next page of results.

        Returns:
            A mapping of URL query parameters.
        """
        params: dict[str, Any] = {}
        marker = self.get_starting_replication_key_value(context)

        if next_page_token:
            params["marker"] = next_page_token
        elif marker:
            params["marker"] = marker
        if self.replication_key:
            params["limit"] = 100
            params["sort"] = "asc"
        if self.book_based and context:
            params["book"] = context["book"]
        return params

    @property
    def partitions(self) -> list[dict[str, Any]] | None:
        """Return a list of partition key dicts (if applicable), otherwise None.

        Returns:
            A list of dictionaries identifying stream partitions.
        """
        if self.book_based:
            return [{"book": book} for book in self.config["books"]]
        return []

    def backoff_max_tries(self) -> int:
        """Return the maximum number of retries for a request.

        Returns:
            The maximum number of retries for a request.
        """
        return 10

    def backoff_wait_generator(self) -> Generator[float, Any, None]:
        """Return a generator of backoff wait times.

        Returns:
            A generator of backoff wait times.
        """
        return backoff.constant(interval=60)

    def validate_response(self, response: requests.Response) -> None:
        """Validate HTTP response.

        By default, checks for error status codes (>400) and raises a
        :class:`singer_sdk.exceptions.FatalAPIError`.

        Tap developers are encouraged to override this method if their APIs use HTTP
        status codes in non-conventional ways, or if they communicate errors
        differently (e.g. in the response body).

        .. image:: ../images/200.png


        In case an error is deemed transient and can be safely retried, then this
        method should raise an :class:`singer_sdk.exceptions.RetriableAPIError`.

        Args:
            response: A `requests.Response`_ object.

        Raises:
            RetriableAPIError: If the request is retriable.

        .. _requests.Response:
            https://docs.python-requests.org/en/latest/api/#requests.Response
        """
        if response.status_code in self.retry_codes:
            raise RetriableAPIError(response.reason)

        super().validate_response(response)



"""Stream type classes for tap-bitso."""

from __future__ import annotations

import sys
import typing as t

from tap_bitso import schemas
from tap_bitso.client import BitsoStream

if sys.version_info >= (3, 9):
    from importlib import resources as importlib_resources
else:
    import importlib_resources

SCHEMAS_DIR = importlib_resources.files(schemas)


class LedgerStream(BitsoStream):
    """Ledger stream.

    DEPRECATED.
    """

    name = "ledger"
    path = "/v3/ledger"
    replication_key = "eid"
    primary_keys: t.ClassVar[list[str]] = ["eid"]
    next_page_token_jsonpath = f"$.payload[-1].{replication_key}"
    schema_filepath = SCHEMAS_DIR / "ledger.json"  # type: ignore[assignment]


class TradesStream(BitsoStream):
    """Trades stream."""

    name = "trades"
    path = "/v3/trades"
    book_based = True
    replication_key = "tid"
    primary_keys: t.ClassVar[list[str]] = ["tid"]
    next_page_token_jsonpath = f"$.payload[-1].{replication_key}"
    schema_filepath = SCHEMAS_DIR / "trade.json"  # type: ignore[assignment]


class UserTradesStream(BitsoStream):
    """User trades stream."""

    name = "user_trades"
    path = "/v3/user_trades"
    book_based = True
    replication_key = "tid"
    primary_keys: t.ClassVar[list[str]] = ["tid"]
    next_page_token_jsonpath = f"$.payload[-1].{replication_key}"
    schema_filepath = SCHEMAS_DIR / "trade.json"  # type: ignore[assignment]


class TickersStream(BitsoStream):
    """Tickers stream."""

    name = "tickers"
    path = "/v3/ticker"
    book_based = True
    records_jsonpath = "$.payload"
    primary_keys: t.ClassVar[list[str]] = ["book", "created_at"]
    schema_filepath = SCHEMAS_DIR / "ticker.json"  # type: ignore[assignment]


class BooksStream(BitsoStream):
    """Books stream."""

    name = "books"
    path = "/v3/available_books"
    primary_keys: t.ClassVar[list[str]] = ["book"]
    schema_filepath = SCHEMAS_DIR / "book.json"  # type: ignore[assignment]



"""Bitso Authentication."""

from __future__ import annotations

import hashlib
import hmac
import logging
import time
from typing import TYPE_CHECKING
from urllib.parse import urlparse

from singer_sdk.authenticators import APIAuthenticatorBase

if TYPE_CHECKING:
    from requests import PreparedRequest

    from tap_bitso.client import BitsoStream

logger = logging.getLogger(__name__)


class BitsoAuthenticator(APIAuthenticatorBase):
    """Authenticator class for Bitso."""

    @classmethod
    def create_for_stream(
        cls: type[BitsoAuthenticator],
        stream: BitsoStream,
    ) -> BitsoAuthenticator:
        """Create the authenticator for the stream.

        Args:
            stream: A Singer stream class.

        Returns:
            An authenticator instance.
        """
        return cls(stream=stream)

    def authenticate_request(
        self: BitsoAuthenticator,
        request: PreparedRequest,
    ) -> PreparedRequest:
        """Modify outgoing request with authentication data.

        See: https://bitso.com/api_info?python#creating-and-signing-requests

        Args:
            request: The `requests.PreparedRequest`_ object.

        Returns:
            The modified `requests.PreparedRequest`_ object.

        .. _requests.Request:
            https://docs.python-requests.org/en/latest/api/#requests.PreparedRequest
        """
        bitso_key: str = self.config["key"]
        bitso_secret: str = self.config["secret"]
        nonce = str(int(round(time.time() * 1000)))

        parsed = urlparse(request.url)
        method = request.method or "GET"
        path = (
            parsed.path if isinstance(parsed.path, str) else parsed.path.decode("utf-8")
        )
        message = nonce + method + path.rstrip("/")

        if method.lower() == "post" and request.body:
            message += (
                request.body
                if isinstance(request.body, str)
                else request.body.decode("utf-8")
            )

        if parsed.query:
            message += "?" + (
                parsed.query
                if isinstance(parsed.query, str)
                else parsed.query.decode("utf-8")
            )

        self.logger.debug("Signing message: %s", message)

        signature = hmac.new(
            bitso_secret.encode("utf-8"),
            message.encode("utf-8"),
            hashlib.sha256,
        ).hexdigest()

        # Update request with Bitso auth
        request.headers["Authorization"] = f"Bitso {bitso_key}:{nonce}:{signature}"

        return request
